@@ -1,8 +1,6 @@
 #ifndef slic3r_ThumbnailData_hpp_
 #define slic3r_ThumbnailData_hpp_
 
-#if ENABLE_THUMBNAIL_GENERATOR
-
 #include <vector>
 #include "libslic3r/Point.hpp"
 
@@ -21,11 +19,19 @@ struct ThumbnailData
     bool is_valid() const;
 };
 
-typedef std::vector<ThumbnailData> ThumbnailsList;
-typedef std::function<void(ThumbnailsList & thumbnails, const Vec2ds & sizes, bool printable_only, bool parts_only, bool show_bed, bool transparent_background)> ThumbnailsGeneratorCallback;
+using ThumbnailsList = std::vector<ThumbnailData>;
+
+struct ThumbnailsParams
+{
+	const Vec2ds 	sizes;
+	bool 			printable_only;
+	bool 			parts_only;
+	bool 			show_bed;
+	bool 			transparent_background;
+};
+
+typedef std::function<ThumbnailsList(const ThumbnailsParams&)> ThumbnailsGeneratorCallback;
 
 } // namespace Slic3r
 
-#endif // ENABLE_THUMBNAIL_GENERATOR
-
-#endif // slic3r_ThumbnailData_hpp_
\ No newline at end of file
+#endif // slic3r_ThumbnailData_hpp_
